/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: itifgIoctl.c,v 7.0.2.1 2000/11/29 15:02:10 cvs Exp $
 *
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

/*
 * Defines:
 */

#undef CHECK_FOR_EIO
#define CHECK_FOR_EIO(function, argument) \
	if (function (iti, argument) < 0) \
	  { \
	    iti_error (iti, # function); \
	    error = -EIO; \
	    goto ioctl_error; \
	  }

/*
 * Globals:
 */

extern struct vs_cam_t vs_def_cam;	/* \               */
extern struct dg_cam_t dg_def_cam;	/* - see itifgAM.c */
extern struct pv_cam_t pv_def_cam;	/* /               */
extern struct pcd_cam_t pcd_def_cam;    /* /		   */

/*
 * Prototypes:
 */

/*
 * iti_ioctl
 */

int
iti_ioctl (iti_file_t *file,
	   iti_dev_t dev,
	   u_int cmd,
	   union iti_inparam_t *inparam,
	   union iti_outparam_t *outparam)
{

  int error, board, funct, pcdig = -1;
  itifg_t *iti;

  u_char exsclk = 0;
  u_long exstime = 0, midpnt = 0;
  double exstime_convert = 0.0, midpnt_convert = 0.0;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  
#if (DEBUG_ITIFG & 1)  
  {
    const struct ioctl_cmd_t 
    {
      u_int number;
      char text[32];
    } 
    ioctl_cmd[] =
    {
      { GIOC_CR_SET_STOP,	"CR_SET_STOP" },
      { GIOC_CR_SET_START,	"CR_SET_START" },
      { GIOC_CR_SET_TRIG,	"CR_SET_TRIG" },
      { GIOC_CR_SET_HDEC,	"CR_SET_HDEC" },
      { GIOC_CR_SET_VDEC,	"CR_SET_VDEC" },
      { GIOC_VS_GET_LUT, 	"VS_GET_LUT" },
      { GIOC_VS_SET_LUT, 	"VS_SET_LUT" },
      { GIOC_VS_SET_LUT_LIN,	"VS_SET_LUT_LIN" },
      { GIOC_VS_GET_NREF, 	"VS_GET_NREF" },
      { GIOC_VS_SET_NREF, 	"VS_SET_NREF" },
      { GIOC_VS_GET_PREF, 	"VS_GET_PREF" },
      { GIOC_VS_SET_PREF, 	"VS_SET_PREF" },
      { GIOC_DG_GET_LUTPG,	"DG_GET_LUTPG" },
      { GIOC_DG_SET_LUTPG,	"DG_SET_LUTPG" },       
      { GIOC_DG_GET_LUT,	"DG_GET_LUT" },         
      { GIOC_DG_SET_LUT,	"DG_SET_LUT" },         
      { GIOC_DG_GET_OLUTPG,	"DG_GET_OLUTPG" },      
      { GIOC_DG_SET_OLUTPG,	"DG_SET_OLUTPG" },      
      { GIOC_DG_GET_OLUT,	"DG_GET_OLUT" },        
      { GIOC_DG_SET_OLUT,	"DG_SET_OLUT" },        
      { GIOC_DG_SET_LUT_LIN,	"DG_SET_LUT_LIN" },        
      { GIOC_DG_SET_NORM_TO8,	"DG_SET_NORM_TO8" },
      { GIOC_DG_SET_NORM_TO16,	"DG_SET_NORM_TO16" },        
      { GIOC_DG_GET_EXTIME,	"DG_GET_EXTIME" },
      { GIOC_DG_SET_EXTIME,	"DG_SET_EXTIME" },        
      { GIOC_DG_GET_CAMMD,	"DG_GET_CAMMD" },
      { GIOC_DG_SET_CAMMD,	"DG_SET_CAMMD" },        
      { GIOC_GE_GET_WIDTH, 	"GE_GET_WIDTH" },
      { GIOC_GE_GET_HEIGHT, 	"GE_GET_HEIGHT" },
      { GIOC_GE_GET_DEPTH, 	"GE_GET_DEPTH" },
      { GIOC_GE_GET_RAWSIZE, 	"GE_GET_RAWSIZE" },
      { GIOC_GE_GET_PAGEDSIZE, 	"GE_GET_PAGEDSIZE" },
      { GIOC_GE_GET_CAMERA, 	"GE_GET_CAMERA" },
      { GIOC_GE_SET_CAMERA, 	"GE_SET_CAMERA" },
      { GIOC_GE_SET_DEFCNF, 	"GE_SET_DEFCNF" },
      { GIOC_GE_GET_CAMCNF, 	"GE_GET_CAMCNF" },
      { GIOC_GE_SET_CAMCNF, 	"GE_SET_CAMCNF" },
      { GIOC_GE_GET_SYNCWAIT,	"GE_GET_SYNCWAIT" },
      { GIOC_GE_SET_TIMEOUT,	"GE_SET_TIMEOUT" },
      { GIOC_GE_GET_STATS,	"GE_GET_STATS" },
      { GIOC_GE_SET_STATS,	"GE_SET_STATS" },
      { HIOC_PV_GET_LUT, 	"PV_GET_LUT" },
      { HIOC_PV_SET_LUT, 	"PV_SET_LUT" },
      { HIOC_PV_SET_LUT_LIN,	"PV_SET_LUT_LIN" },
      { HIOC_PV_GET_NREF, 	"PV_GET_NREF" },
      { HIOC_PV_SET_NREF, 	"PV_SET_NREF" },
      { HIOC_PV_GET_PREF, 	"PV_GET_PREF" },
      { HIOC_PV_SET_PREF, 	"PV_SET_PREF" },
      { HIOC_PCD_GET_LUT0PG,	"PCD_GET_LUT0PG" },
      { HIOC_PCD_SET_LUT0PG,	"PCD_SET_LUT0PG" },       
      { HIOC_PCD_GET_LUT0,	"PCD_GET_LUT0" },         
      { HIOC_PCD_SET_LUT0,	"PCD_SET_LUT0" },         
      { HIOC_PCD_GET_LUT1PG,	"PCD_GET_LUT1PG" },      
      { HIOC_PCD_SET_LUT1PG,	"PCD_SET_LUT1PG" },      
      { HIOC_PCD_GET_LUT1,	"PCD_GET_LUT1" },        
      { HIOC_PCD_SET_LUT1,	"PCD_SET_LUT1" },        
      { HIOC_PCD_SET_LUT_LIN,	"PCD_SET_LUT_LIN" },        
      { HIOC_PCD_SET_NORM_TO8,	"PCD_SET_NORM_TO8" },
      { HIOC_PCD_SET_NORM_TO16,	"PCD_SET_NORM_TO16" },
      { HIOC_PCD_SET_EXFREE,	"PCD_SET_EXFREE" },        
      { HIOC_PCD_GET_EXTIME,	"PCD_GET_EXTIME" },
      { HIOC_PCD_SET_EXTIME,	"PCD_SET_EXTIME" },        
      { HIOC_PCD_GET_CAMMD,	"PCD_GET_CAMMD" },
      { HIOC_PCD_SET_CAMMD,	"PCD_SET_CAMMD" },
      { HIOC_PCD_GET_STMODE,	"PCD_GET_STMODE" },
      { HIOC_PCD_SET_STMODE,	"PCD_SET_STMODE" },
      { HIOC_PCD_GET_ETMODE,	"PCD_GET_ETMODE" },
      { HIOC_PCD_SET_ETMODE,	"PCD_SET_ETMODE" },
      { HIOC_PCD_GET_FCNT,	"PCD_GET_FCNT" },
      { HIOC_PCD_SET_FCNT,	"PCD_SET_FCNT" },
      { 0, " " }
    };

    char cmd_text[32];
    int  index = 0;

    iti_strcpy (cmd_text, "COMMAND NOT SUPPORTED");
    while (ioctl_cmd[index].number != 0)
      {
	if (ioctl_cmd[index].number == cmd)
	  {
	    iti_strcpy (cmd_text, ioctl_cmd[index].text);
	    break;
	  }
	index++;
      }

  ITI_PRINT0A("> iti_ioctl %d: %s\n", board, cmd_text);
  ITI_PRINT0A("    in 0x%p, out 0x%p\n", inparam, outparam);
  }
#endif

  iti = itifg[board];
 
  /* test for configured hardware */
  if (!FPGA_OK)
    {
      error = -ENXIO;
      goto ioctl_error;
    }

  /* only some ioctl's for the io-port devices yet */
  if (!IS_DATA)
    {
      if (cmd != GIOC_GE_GET_STATS && cmd != GIOC_GE_SET_STATS)
	{
	  error = -ENXIO;
	  goto ioctl_error;
	}
    }

  /* which calls accept a 'NULL' argument */
  switch (cmd)
    {
    case GIOC_GE_GET_SYNCWAIT:
      break;
    default:
      if (!inparam && !outparam)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
    }

  /* test for image acqusition */
  switch (cmd)
    {
    case GIOC_CR_SET_HDEC:
    case GIOC_CR_SET_VDEC:
    case GIOC_DG_SET_NORM_TO8:
    case GIOC_DG_SET_NORM_TO16:
    case GIOC_GE_SET_CAMERA:
    case GIOC_GE_SET_DEFCNF:
    case GIOC_GE_SET_CAMCNF:
    case GIOC_GE_GET_SYNCWAIT:
    case HIOC_PCD_SET_NORM_TO8:
    case HIOC_PCD_SET_NORM_TO16:
    case HIOC_PCD_SET_EXFREE:
      if (iti->opmode & ITI_WAIT_MASK)
	{
	  ITI_PRINT4("    opmode: WAIT\n");
	  error = -EBUSY;
	  goto ioctl_error;
	}
      ITI_PRINT4("    opmode: !WAIT\n");
    }

  switch (cmd)
    {
      /* GENERIC */
      
      /* get width of buffer */
    case GIOC_GE_GET_WIDTH:
      outparam->get_width = iti->width;
      break;
      
      /* get height of buffer */
    case GIOC_GE_GET_HEIGHT:
      outparam->get_height = iti->height;
      break;

      /* depth of buffer */
    case GIOC_GE_GET_DEPTH:
      outparam->get_depth = iti->get_pixsz (iti);
      outparam->get_depth *= 8;
      switch (MODULE_ID)
	{
	case ICP_AMVS:
	  break;
	case ICP_AMDG:
	  if (outparam->get_depth == 16)
	    outparam->get_depth = dg_get_psize (iti);
	  break;
	case ICP_AMPV:
	  break;
	case ITI_VMPCD:
	  if (outparam->get_depth == 16)
	    outparam->get_depth = pcd_get_lut0psz (iti);
	  break;
	default:
	  error = -ENXIO;
	  goto ioctl_error;
	}
      break;

      /* get 'raw' size (width * height * depth) of every buffer */
    case GIOC_GE_GET_RAWSIZE:
      outparam->get_rawsize = iti->raw_size;
      break;

      /* get 'paged' size ('raw' size + iti_info_t aligned) of every buffer */
    case GIOC_GE_GET_PAGEDSIZE:
      outparam->get_pagedsize = iti->paged_size;
      break;

      /* get current camera */
    case GIOC_GE_GET_CAMERA:
      outparam->get_camera = iti->cur_cam;
      break;

      /* set current camera */ 
    case GIOC_GE_SET_CAMERA:
      ITI_PRINT1A("    camera: %d\n", inparam->set_camera);
      /* is the camera number valid? */
      if (inparam->set_camera > iti->cameras)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      /* no switching neccessary at digital boards */
      if (IS_AMDG || IS_VMPCD)
	goto ioctl_done;
      /* stop image acquisition (synchron) */
      CHECK_FOR_EIO (iti->set_acqmd, CR_FREEZE);
      iti->opmode |= ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (|SYNC)\n", iti->opmode);
      ITI_PRINT4("    status: sleep\n");
      if ((error = iti_sleep (iti, iti->cam_expire, ITI_CAMERA)))
	goto ioctl_error;
      iti->opmode &= ~ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (&~SYNC)\n", iti->opmode);
      /* switch input port */
      switch (MODULE_ID)
	{
	case ICP_AMVS:
	  CHECK_FOR_EIO (vs_set_camsel, inparam->set_camera);
	  break;
	case ICP_AMPV:
	  CHECK_FOR_EIO (pv_set_camsel, inparam->set_camera);
	  break;
	default:
	  error = -ENXIO;
	  goto ioctl_error;
	}
      /* update structure */
      iti->cur_cam = inparam->set_camera;
      /* start image acqusition (synchron) */
      CHECK_FOR_EIO (iti->set_acqmd, CR_GRAB);
      iti->opmode |= ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (|SYNC)\n", iti->opmode);
      ITI_PRINT4("    status: sleep\n");
      if ((error = iti_sleep (iti, iti->cam_expire, ITI_CAMERA)))
	goto ioctl_error;
      iti->opmode &= ~ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (&~SYNC)\n", iti->opmode);
      /* reset frame rate detection times */
      iti_memset (&iti->thistime, 0x00, sizeof(struct timeval));
      iti_memset (&iti->lasttime, 0x00, sizeof(struct timeval));
      break;

      /* get camera structure */
    case GIOC_GE_GET_CAMCNF:
      switch (MODULE_ID)
	{
	case ICP_AMVS:
	  iti_memcpy (&outparam->get_iti_camcnf, &iti->dev.vs_cam,
		      sizeof(struct vs_cam_t));
	  break;
	case ICP_AMDG:
	  iti_memcpy (&outparam->get_iti_camcnf.dg_cam,
		      &iti->dev.dg_cam, sizeof(struct dg_cam_t));
	  break;
	case ICP_AMPV:
	  iti_memcpy (&outparam->get_iti_camcnf, &iti->dev.pv_cam,
		      sizeof(struct pv_cam_t));
	  iti_memcpy (&outparam->get_iti_camcnf.pv_cam,
		      &iti->dev.pv_cam, sizeof(struct pv_cam_t));
	  break;
	case ITI_VMPCD:
	  iti_memcpy (&outparam->get_iti_camcnf.pcd_cam,
		      &iti->dev.pcd_cam, sizeof(struct pcd_cam_t));
	  break;
	default:
	  error = -ENXIO;
	  goto ioctl_error;
	}
      break;
      
      /* set camera structure */
    case GIOC_GE_SET_DEFCNF:
    case GIOC_GE_SET_CAMCNF:
      switch (MODULE_ID)
	{
	case ICP_AMVS:
	  if (cmd == GIOC_GE_SET_CAMCNF)
	    iti_memcpy (&iti->dev.vs_cam,
			&inparam->set_iti_camcnf.vs_cam,
			sizeof(struct vs_cam_t));
	  else
	    iti_memcpy (&iti->dev.vs_cam, &vs_def_cam,
			sizeof(struct vs_cam_t));
	  if (icp_amvs_init (iti))
	    {
	      error = -EIO;
	      goto ioctl_error;
	    }
	  break;
	case ICP_AMDG:
	  if (cmd == GIOC_GE_SET_CAMCNF)
	    iti_memcpy (&iti->dev.dg_cam,
			&inparam->set_iti_camcnf.dg_cam,
			sizeof(struct dg_cam_t));
	  else
	    iti_memcpy (&iti->dev.dg_cam, &dg_def_cam,
			sizeof(struct dg_cam_t));
	  if (icp_amdg_init (iti))
	    {
	      error = -EIO;
	      goto ioctl_error;
	    }
	  switch (dg_get_psize (iti))
	    {
	    case 8:
	      CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_8);
	      break;
	    case 10:
	    case 12:
	      CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_16);
	    case 16:
	      if (dg_get_ilpixel (iti) == DG_ILEAVE)
		{
		  CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_8);
		}
	      else
		{
		  CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_16);
		}
	      break;
	    default:
	      error = -EIO;
	      goto ioctl_error;
	    }
	  break;
	case ICP_AMPV:
	  if (cmd == GIOC_GE_SET_CAMCNF)
	    iti_memcpy (&iti->dev.pv_cam,
			&inparam->set_iti_camcnf.pv_cam,
			sizeof(struct pv_cam_t));
	  else
	    iti_memcpy (&iti->dev.pv_cam, &pv_def_cam,
			sizeof(struct pv_cam_t));
	  if (icp_ampv_init (iti))
	    {
	      error = -EIO;
	      goto ioctl_error;
	    }
	  break;
	case ITI_VMPCD:
	  if (cmd == GIOC_GE_SET_CAMCNF)
	    iti_memcpy (&iti->dev.pcd_cam,
			&inparam->set_iti_camcnf.pcd_cam,
			sizeof(struct pcd_cam_t));
	  else
	    iti_memcpy (&iti->dev.pcd_cam, &pcd_def_cam,
			sizeof(struct pcd_cam_t));
	  if (pcd_viam_init (iti))
	    {
	      error = -EIO;
	      goto ioctl_error;
	    }
	  break;
	default:
	  error = -ENXIO;
	  goto ioctl_error;
	}
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      /* reenable the right vsync irq */
      if ((error = iti_enable_irq (iti, ITI_EIRQ_DATA_OPEN)))
	goto ioctl_error;
      break;
      
      /* get the camera sync impulse period (or only wait for sync) */
    case GIOC_GE_GET_SYNCWAIT:
      iti->opmode |= ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (|SYNC)\n", iti->opmode);
      ITI_PRINT4("    status: sleep\n");
      if ((error = iti_sleep (iti, iti->cam_expire, ITI_CAMERA)))
	{
	  iti->opmode &= ~ITI_WAIT_SYNC;
	  ITI_PRINT4A("    opmode: %#x (&~SYNC)\n", iti->opmode);
	  goto ioctl_error;
	}
      if (iti->opmode & ITI_TIMEOUT)
	{
	  ITI_PRINT4("    opmode: TIME\n");
	  iti->opmode &= ~ITI_WAIT_SYNC;
	  iti->opmode &= ~ITI_TIMEOUT;
	  ITI_PRINT4A("    opmode: %#x (&~SYNC&~TIME)\n",
		      iti->opmode);  
	  error = -ENODATA;
	  goto ioctl_error;
	}
      ITI_PRINT4("    opmode: !TIME\n");
      if (outparam)
	{
	  struct timeval difftime;
	  
	  ITI_PRINT4("    status: sleep\n");
	  if ((error = iti_sleep (iti, iti->cam_expire, ITI_CAMERA)))
	    {
	      iti->opmode &= ~ITI_WAIT_SYNC;
	      ITI_PRINT4A("    opmode: %#x (&~SYNC)\n",
			  iti->opmode);
	      goto ioctl_error;
	    }
	  if (iti->opmode & ITI_TIMEOUT)
	    {
	      ITI_PRINT4("    opmode: TIME\n");
	      iti->opmode &= ~ITI_WAIT_SYNC;
	      iti->opmode &= ~ITI_TIMEOUT;
	      ITI_PRINT4A("    opmode: %#x (&~SYNC&~TIME)\n",
			  iti->opmode);  
	      error = -ENODATA;
	      goto ioctl_error;
	    }
	  ITI_PRINT4("    opmode: !TIME\n");
	  TV_MINUS(difftime, iti->thistime, iti->lasttime);
	  iti_memcpy (&outparam->get_syncrate, &difftime,
		      sizeof(struct timeval));
	}
      iti->opmode &= ~ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (&~SYNC)\n", iti->opmode);
      break;
      
      /* set the image acqusition timeout value */
    case GIOC_GE_SET_TIMEOUT:
      {
	u_long tmp;
	time_t new;
	
	TV_TO_MS (inparam->set_timeout, tmp);
	ITI_PRINT1A("    timeout: %ldms\n", tmp);
	new = (tmp * HZ) / 1000;
	if ((new * 1000) / HZ < tmp)
	  new++; /* round up */
	new = (time_t)((float)new * 1.25);
	ITI_PRINT1A("    expire: %ld\n", new);
	if (new < 1 || new > 10000)
	  {
	    error = -EINVAL;
	    goto ioctl_error;
	  }
	iti->cam_expire = new;
      }
    break;
    
      /* get some statistics */
    case GIOC_GE_GET_STATS:
      iti_memcpy (&outparam->get_stats, &iti->acc,
		  sizeof (struct iti_acc_t));
      break;
	
      /* reset statistics */
    case GIOC_GE_SET_STATS:
      iti->acc.events_captured = 0UL;
      iti->acc.bytes_valid = 0UL;
      iti->acc.bytes_lost = 0UL;
      break;
      
      /* IC-PCI */
      
      /* stop online video */
    case GIOC_CR_SET_STOP:
      iti->opmode &= ~ITI_MODE_GRAB;
      iti->opmode &= ~ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (&~GRAB&~SYNC)\n", iti->opmode);
      iti->started = FALSE;
      /* delete the vsync test timer */
      iti_del_timer (iti, ITI_CAMERA);
      /* if grab is stopped, any initialized vectors are invalid! */
      iti_clear_sglist (iti);
      break;
      
      /* start online video */
    case GIOC_CR_SET_START:
      /* online video is already started */
      if (iti->opmode & ITI_MODE_GRAB)
	{
	  ITI_PRINT4("    opmode: GRAB\n");
	  goto ioctl_done;
	}
      ITI_PRINT4("    opmode: !GRAB\n");
      /* has any other driver activation successfully finished? */
      if (iti->opmode & ITI_WAIT_MASK)
	{
	  ITI_PRINT4("    opmode: WAIT\n");
	  error = -EBUSY;
	  goto ioctl_error;
	}
      ITI_PRINT4("    opmode: !WAIT\n");
      /* is the grab buffer available at all? */
      if (!iti->km_size)
	{
	  error = -EAGAIN;
	  goto ioctl_error;
	}
      /* is the buffer big enough to hold at least one image? */
      if (iti->km_size < iti->paged_size)
	{
	  error = -ENOMEM;
	  goto ioctl_error;
	}
      /* set source and destination scatter gather tables */
      if ((error = iti_src_init (iti)))
	goto ioctl_error;
      if ((error = iti_dst_init (iti, iti->km_base,
				 (iti->km_size /
				 iti->paged_size) *
				 iti->paged_size, FALSE)))
	goto ioctl_error;
      /* start the contiguous image acquistion */
      iti_irq_lock (iti);
      ZERO_ACNT(cam);
      ZERO_ACNT(blk);
      iti->set_intadr (iti, iti->offset *
				  ITI_BLOCK_SIZE);
      iti->enabled = FALSE;
      if (IS_PCDIG)
	iti->set_adrien (iti, CR_ENABLE);
      /* if an acq. is already running, delay this until the next vsync */
      if (iti->get_aqadr (iti))
	iti->started = FALSE;
      else
	iti->started = TRUE;
      ZERO_ACNT(iov);
      iti->opmode |= ITI_MODE_GRAB;
      ITI_PRINT4A("    opmode: %#x (|GRAB)\n", iti->opmode);
      iti_irq_unlock (iti);
      /* add the vsync test timer */      
      iti_add_timer (iti, iti->cam_expire * 3, ITI_CAMERA);
      ITI_PRINT4("    status: timer\n");
      /* update the current frame count for select */
      iti->frames_selected = iti->acc.frames_transfered;
      break;
      
      /* trigger per software */
    case GIOC_CR_SET_TRIG:
      switch (MODULE_ID)
	{
	case ICP_AMVS:
	  /* test for trigger mode */
	  if (vs_get_trigen (iti) != VS_ENABLE)
	    { 
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  /* test for trigger source */
	  if (vs_get_trigsrc (iti) != VS_TRIG_SOFT)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  /* trigger board */
	  if (vs_sw_trig (iti))
	    {
	      iti_error (iti, "vs_sw_trig");
	      error = -EIO;
	      goto ioctl_error;
	    }
	  break;
	case ICP_AMDG:
	  /* test for trigger mode */
	  if (dg_get_extmd (iti) != DG_SOFT_TRIG)
	    {
	      if (dg_get_trigen (iti) != DG_ENABLE)
		{ 
		  error = -EINVAL;
		  goto ioctl_error;
		}
	      /* test for trigger source */
	      if (dg_get_trigsrc (iti) != DG_TRIG_SOFT)
		{
		  error = -EINVAL;
		  goto ioctl_error;
		}
	    }
	  /* trigger */
	  if (dg_sw_trig (iti))
	    { 
	      iti_error (iti, "dg_sw_trig");
	      error = -EIO;
	      goto ioctl_error;
	    }
	  break;
	case ICP_AMPV:
	  /* test for trigger mode */
	  if (pv_get_trigen (iti) != PV_ENABLE)
	    { 
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  /* test for trigger source */
	  if (pv_get_trigsrc (iti) != PV_TRIG_SOFT)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  /* trigger board */
	  if (pv_sw_trig (iti))
	    { 
	      iti_error (iti, "pv_sw_trig");
	      error = -EIO;
	      goto ioctl_error;
	    }
	  break;
	case ITI_VMPCD:
	  /* find first board */
	  for (pcdig = board; pcdig >= 0; pcdig--)
	    if (itifg[pcdig]->device_id == ITI_PCDIG &&
		pcd_get_driveext (itifg[pcdig]) == PCD_ENABLE)
	      break;
	  if (pcdig < 0)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	      }
	  /* test for trigger mode enabled */
	  if (pcd_get_exsmd (itifg[pcdig]) != PCD_EXSYNC_SW_TRIG0 &&
	      pcd_get_stmode (itifg[pcdig]) != PCD_TMODE_SWT0)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  /* trigger first board */
	  if (pcd_sw_trig0 (itifg[pcdig]))
	    { 
	      iti_error (itifg[pcdig], "pcd_sw_trig0");
	      error = -EIO;
		goto ioctl_error;
	    }
	  break;
	default:
	  error = -ENXIO;
	  goto ioctl_error;
	}
      break;

      /* set horizontal decimation; (clipping/padding) */
    case GIOC_CR_SET_HDEC:
      ITI_PRINT1A("    horiz. decimation: %d\n", inparam->set_hdec);
      if (IS_REV_A && inparam->set_hdec != 1)
	{
	  /* old boards */
	  error = -ENXIO;
	  goto ioctl_error;
	}
      /* new boards */
      switch (inparam->set_hdec)
	{
	case 1:
	  inparam->set_hdec = CR_DEC_BY_1;
	  break;	  
	case 2:
	  inparam->set_hdec = CR_DEC_BY_2;
	  break;	  
	case 4:
	  inparam->set_hdec = CR_DEC_BY_4;
	  break;	  
	case 8:
	  inparam->set_hdec = CR_DEC_BY_8;
	  if (IS_ICPCI && iti->get_pixsz (iti) != 1)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  break;
	case 16:
	  inparam->set_hdec = CR_DEC_BY_16;
	  if (IS_ICPCI)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  break;
	default:
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (iti->set_hdec, inparam->set_hdec);
      /* recalculate buffer size */
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      break;

      /* set vertical decimation; (clipping/padding) */
    case GIOC_CR_SET_VDEC:
      ITI_PRINT1A("    vert. decimation: %d\n", inparam->set_vdec);
      if (IS_REV_A && inparam->set_vdec != 1)
	{
	  /* old boards */
	  error = -ENXIO;
	  goto ioctl_error;
	}
      /* new boards */
      switch (inparam->set_vdec)
	{
	case 1:
	  inparam->set_vdec = CR_DEC_BY_1;
	  break;	  
	case 2:
	  inparam->set_vdec = CR_DEC_BY_2;
	  break;	  
	case 4:
	  inparam->set_vdec = CR_DEC_BY_4;
	  break;	  
	case 8:
	  inparam->set_vdec = CR_DEC_BY_8;
	  break;
	case 16:
	  if (IS_ICPCI)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_vdec = CR_DEC_BY_16;
	  break;
	default:
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (iti->set_vdec, inparam->set_vdec);
      /* recalculate buffer size */
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      /* set vsync irq by current vertical decimation */
      if (iti->get_smstat (iti) == CR_ILACE &&
	  iti->get_vdec (iti) == CR_DEC_BY_1)
	{
	  CHECK_FOR_EIO (iti->set_fldsel, CR_EVEN_FIELD);
	}
      else
	{
	  CHECK_FOR_EIO (iti->set_fldsel, CR_NEXT_FIELD);
	}
      /* reenable the right vsync irq */
      if ((error = iti_enable_irq (iti, ITI_EIRQ_DATA_OPEN)))
	goto ioctl_error;
      break;

      /* AM-VS */
      
      /* read lookup table (256 values) */
    case GIOC_VS_GET_LUT:
      vs_get_lutdata (iti, &outparam->get_vs_lut);
      break;
      
      /* write lookup table (256 values) */
    case GIOC_VS_SET_LUT:
      CHECK_FOR_EIO (vs_set_lutdata, &inparam->set_vs_lut);
      break;

      /* linearize lookup table */
    case GIOC_VS_SET_LUT_LIN:
      if (vs_lin_lut (iti))
	{
	  iti_error (iti, "vs_lin_lut");
	  error = -EIO;
	  goto ioctl_error;
	}
      break;

      /* read negative reference voltage */
    case GIOC_VS_GET_NREF:
      outparam->get_vs_nref = vs_get_nref (iti); 
      break;

      /* write negative reference voltage */
    case GIOC_VS_SET_NREF:
      ITI_PRINT1A("    negative-ref : %d\n", inparam->set_vs_nref);
      if (inparam->set_vs_nref > VS_NREF_MAX)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (vs_set_nref, inparam->set_vs_nref);
      break;

      /* read positive reference voltage */
    case GIOC_VS_GET_PREF:
      outparam->get_vs_pref = vs_get_pref (iti); 
      break;

      /* write positive reference voltage */
    case GIOC_VS_SET_PREF:
      ITI_PRINT1A("    positive-ref : %d\n", inparam->set_vs_pref);
      if (inparam->set_vs_pref > VS_PREF_MAX)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (vs_set_pref, inparam->set_vs_pref);
      break;

      /* AM-DIG */
      
      /* read 16 bit input LUT page number */
    case GIOC_DG_GET_LUTPG:
      outparam->get_dg_lutpg = dg_get_lutpg (iti); 
      switch (dg_get_psize (iti))
	{
	case 16:
	  outparam->get_dg_lutpg >>= DG_PSIZE_16_SHIFT;
	  break;
	case 12:
	  outparam->get_dg_lutpg >>= DG_PSIZE_12_SHIFT;
	  break;
	case 10:
	  outparam->get_dg_lutpg >>= DG_PSIZE_10_SHIFT;
	  break;
	case 8:
	  outparam->get_dg_lutpg >>= DG_PSIZE_8_SHIFT;
	  break;
	}
      break;

      /* write 16 bit input LUT page number */
    case GIOC_DG_SET_LUTPG:
      switch (dg_get_psize (iti))
	{
	case 16:
	  if (inparam->set_dg_lutpg & ~DG_PSIZE_16_MASK)
 	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_dg_lutpg <<= DG_PSIZE_16_SHIFT;
	  break;
	case 12:
	  if (inparam->set_dg_lutpg & ~DG_PSIZE_12_MASK)
 	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_dg_lutpg <<= DG_PSIZE_12_SHIFT;
	  break;
	case 10:
	  if (inparam->set_dg_lutpg & ~DG_PSIZE_10_MASK)
 	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_dg_lutpg <<= DG_PSIZE_10_SHIFT;
	  break;
	case 8:
	  if (inparam->set_dg_lutpg & ~DG_PSIZE_8_MASK)
 	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_dg_lutpg <<= DG_PSIZE_8_SHIFT;
	  break;
	}
      CHECK_FOR_EIO (dg_set_lutpg, inparam->set_dg_lutpg);
      break;

      /* read 16 bit input lookup table */
    case GIOC_DG_GET_LUT:
      dg_get_lut16 (iti, '0', &outparam->get_dg_lut16);
      break;
      
      /* write 16 bit input lookup table */
    case GIOC_DG_SET_LUT:
      if (dg_set_lut16 (iti, '0', &inparam->set_dg_lut16))
	{
	  iti_error (iti, "dg_set_lut16");
	  error = -EIO;
	  goto ioctl_error;
	}
      break;
      
      /* read 8 bit optional input LUT page number */
    case GIOC_DG_GET_OLUTPG:
      outparam->get_dg_olutpg = dg_get_olutpg (iti); 
      break;
      
      /* write 8 bit optional input LUT page number */
    case GIOC_DG_SET_OLUTPG:
      CHECK_FOR_EIO (dg_set_olutpg, inparam->set_dg_olutpg);
      break;
      
      /* read 8 bit optional input lookup table */
    case GIOC_DG_GET_OLUT:
      dg_get_lut8 (iti, &outparam->get_dg_lut8);
      break;

      /* write 8 bit optional input lookup table */
    case GIOC_DG_SET_OLUT:
      CHECK_FOR_EIO (dg_set_lut8, &inparam->set_dg_lut8);
      break;
		     
      /* linearize lookup table */
    case GIOC_DG_SET_LUT_LIN:
      if (dg_get_modver (iti) != DG_MODVER_8_TTL &&
	  dg_get_modver (iti) != DG_MODVER_8_RS422)
        {
	  if (dg_lin_lut16 (iti, '0', ITI_MODE_LINEAR))
	    {
	      error = -EIO;
	      goto ioctl_error;
	    }
	}
      if (dg_get_modver (iti) == DG_MODVER_24_TTL)
	if (dg_lin_lut8 (iti) < 0)
	  {
	    iti_error (iti, "dg_lin_lut8");
	    return -EIO;
	  }
      switch (dg_get_modver (iti))
	{
	case DG_MODVER_8_TTL:
	case DG_MODVER_8_RS422:
	  CHECK_FOR_EIO (dg_set_muxb, DG_MUXB_LO);
	  CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_8);
	  break;
	case DG_MODVER_16_TTL:
	case DG_MODVER_16_RS422:
	  CHECK_FOR_EIO (dg_set_muxb, DG_MUXB_HI);
	  CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_16);
	  break;
	case DG_MODVER_24_TTL:
	  CHECK_FOR_EIO (dg_set_muxb, DG_MUXB_HI);
	  if (IS_REV_B)
	    {
	      CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_24_24);
	    }
	  else
	    {
	      CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_24_32);
	    }
	  break;
	default:
	  return -EIO;
	}
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      break;
      
      /* normalize output to 8 bit (from 8, 10, 12, 16 bit input) */
    case GIOC_DG_SET_NORM_TO8:
      if (dg_get_modver (iti) != DG_MODVER_8_TTL &&
	  dg_get_modver (iti) != DG_MODVER_8_RS422)
	{
	  if (dg_lin_lut16 (iti, '0', ITI_MODE_NORM_TO8))
	    {
	      iti_error (iti, "dg_lin_lut16");
	      error = -EIO;
	      goto ioctl_error;
	    }
	}
      /* the second channel is invalid */
      CHECK_FOR_EIO (dg_set_muxb, DG_MUXB_LO);
      /* set pixelsize an recalculate buffersizes */
      CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_8);
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      break;

      /* normalize output to 16 bit (from 8, 10, 12, 16 bit input) */
    case GIOC_DG_SET_NORM_TO16:
      if (dg_get_modver (iti) != DG_MODVER_8_TTL &&
	  dg_get_modver (iti) != DG_MODVER_8_RS422)
	{
	  if (dg_lin_lut16 (iti, '0', ITI_MODE_NORM_TO16))
	    {
	      iti_error (iti, "dg_lin_lut16");
	      error = -EIO;
	      goto ioctl_error;
	    }
	}
      else
	{
	  error = -EIO;
	  goto ioctl_error;
	}
      /* the second channel is valid */
      CHECK_FOR_EIO (dg_set_muxb, DG_MUXB_HI);
      /* set pixelsize an recalculate buffersizes */
      CHECK_FOR_EIO (cr_set_pixsz, CR_PIXSZ_16);
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      break;

      /* get the exposure time */
    case GIOC_DG_GET_EXTIME:
      exsclk = dg_get_exclk (iti);
      exstime = dg_get_extime (iti);
      switch (exsclk)
	{
	case DG_C_7_k:
	  exstime <<= 11;
	  break;
	case DG_C_28_k:
	  exstime <<= 9;
	  break;
	case DG_C_112_k:
	  exstime <<= 7;
	  break;
	case DG_C_447_5_k:
	  exstime <<= 5;
	  break;
	case DG_C_1_79_M:
	  exstime <<= 3;
	  break;
	case DG_C_7_16_M:
	  exstime <<= 1;
	  break;
	default:
	  error = -EIO;
	  goto ioctl_error;
	}
      exstime *= ITI_DLY_BASE;
      outparam->get_dg_extime.tv_sec = exstime / 1000000000;
      outparam->get_dg_extime.tv_nsec = exstime % 1000000000;
      break;

      /* set the exposure time */
    case GIOC_DG_SET_EXTIME:
      ITI_PRINT1A("    shuttertime : %lds/%ldns\n",
		  inparam->set_dg_extime.tv_sec,
		  inparam->set_dg_extime.tv_nsec );
      exstime = inparam->set_dg_extime.tv_sec +
	inparam->set_dg_extime.tv_nsec;
      if (exstime > DG_DLY_MAX || exstime < DG_DLY_MIN)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      exstime /= ITI_DLY_BASE;
      if (exstime > (1 << 9) * 4098)
	exsclk = DG_C_7_k, exstime >>= 11; 
      else
	if (exstime > (1 << 7) * 4098)
	  exsclk = DG_C_28_k, exstime >>= 9; 
	else
	  if (exstime > (1 << 5) * 4098)
	    exsclk = DG_C_112_k, exstime >>= 7; 
	  else
	    if (exstime > (1 << 3) * 4098)
	      exsclk = DG_C_447_5_k, exstime >>= 5; 
	    else
	      if (exstime > (1 << 1) * 4098)
		exsclk = DG_C_1_79_M, exstime >>= 3; 
	      else
		exsclk = DG_C_7_16_M, exstime >>= 1; 
      CHECK_FOR_EIO (dg_set_exclk, exsclk);
      CHECK_FOR_EIO (dg_set_extime, exstime);
      break;

      /* get camera mode */
    case GIOC_DG_GET_CAMMD:
      outparam->get_dg_cammd = dg_get_mc (iti);
      break;

      /* set camera mode */
    case GIOC_DG_SET_CAMMD:
      ITI_PRINT1A("    cameramode : %d\n", inparam->set_dg_cammd);
      if (inparam->set_dg_cammd < DG_MC_MIN ||
	  inparam->set_dg_cammd > DG_MC_MAX)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (dg_set_mc, inparam->set_dg_cammd);
      break;
      
      /* PCVision */

      /* read lookup table (256 values) */
    case HIOC_PV_GET_LUT:
      pv_get_lutdata (iti, &outparam->get_pv_lut);
      break;

      /* write lookup table (256 values) */
    case HIOC_PV_SET_LUT:
      CHECK_FOR_EIO (pv_set_lutdata, &inparam->set_pv_lut);
      break;

      /* linearize lookup table */
    case HIOC_PV_SET_LUT_LIN:
      if (pv_lin_lut (iti))
	{
	  iti_error (iti, "pv_lin_lut");
	  error = -EIO;
	  goto ioctl_error;
	}
      break;

      /* read negative reference voltage */
    case HIOC_PV_GET_NREF:
      outparam->get_pv_nref = pv_get_nref (iti); 
      break;

      /* write negative reference voltage */
    case HIOC_PV_SET_NREF:
      ITI_PRINT1A("    negative-ref : %d\n", inparam->set_pv_nref);
      if (inparam->set_pv_nref > PV_MAX_NREF)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (pv_set_nref, inparam->set_pv_nref);
      break;

      /* read positive reference voltage */
    case HIOC_PV_GET_PREF:
      outparam->get_pv_pref = pv_get_pref (iti); 
      break;

      /* write positive reference voltage */
    case HIOC_PV_SET_PREF:
      ITI_PRINT1A("    positive-ref : %d.\n", inparam->set_pv_pref);
      if (inparam->set_pv_pref > PV_MAX_PREF)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (pv_set_pref, inparam->set_pv_pref);
      break;

      /* PCDig */

      /* read 16 bit input LUT0 (bit 0-15) page number */
    case HIOC_PCD_GET_LUT0PG:
      outparam->get_pcd_lut0pg = pcd_get_lut0pg (iti); 
      switch (pcd_get_lut0psz (iti))
	{
	case 16:
	  outparam->get_pcd_lut0pg >>= PCD_PSIZE_16_SHIFT;
	  break;
	case 14:
	  outparam->get_pcd_lut0pg >>= PCD_PSIZE_16_SHIFT;
	  break;
	case 12:
	  outparam->get_pcd_lut0pg >>= PCD_PSIZE_12_SHIFT;
	  break;
	case 10:
	  outparam->get_pcd_lut0pg >>= PCD_PSIZE_10_SHIFT;
	  break;
	case 8:
	  outparam->get_pcd_lut0pg >>= PCD_PSIZE_8_SHIFT;
	  break;
	}
      break;

      /* write 16 bit input LUT0 (bit 0-15) page number */
    case HIOC_PCD_SET_LUT0PG:
      switch (pcd_get_lut0psz (iti))
	{
	case 16:
	  if (inparam->set_pcd_lut0pg & ~PCD_PSIZE_16_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut0pg <<= PCD_PSIZE_16_SHIFT;
	  break;
	case 14:
	  if (inparam->set_pcd_lut0pg & ~PCD_PSIZE_14_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut0pg <<= PCD_PSIZE_16_SHIFT;
	  break;
	case 12:
	  if (inparam->set_pcd_lut0pg & ~PCD_PSIZE_12_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut0pg <<= PCD_PSIZE_12_SHIFT;
	  break;
	case 10:
	  if (inparam->set_pcd_lut0pg & ~PCD_PSIZE_10_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut0pg <<= PCD_PSIZE_10_SHIFT;
	  break;
	case 8:
	  if (inparam->set_pcd_lut0pg & ~PCD_PSIZE_8_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut0pg <<= PCD_PSIZE_8_SHIFT;
	  break;
	}
      CHECK_FOR_EIO (pcd_set_lut0pg, inparam->set_pcd_lut0pg);
      break;

      /* read 16 bit input LUT0 (bit 0-15) */
    case HIOC_PCD_GET_LUT0:
      pcd_get_lut (iti, '0', &outparam->get_pcd_lut0);
      break;

      /* write 16 bit input LUT0 (bit 0-15) */
    case HIOC_PCD_SET_LUT0:
      if (pcd_set_lut (iti, '0', &inparam->set_pcd_lut0))
	{
	  iti_error (iti, "pcd_set_lut0");
	  error = -EIO;
	  goto ioctl_error;
	}
      break;
      
      /* read 16 bit input LUT1 (bit 16-31) page number */
    case HIOC_PCD_GET_LUT1PG:
      outparam->get_pcd_lut1pg = pcd_get_lut1pg (iti); 
      switch (pcd_get_lut1psz (iti))
	{
	case 16:
	  outparam->get_pcd_lut1pg >>= PCD_PSIZE_16_SHIFT;
	  break;
	case 14:
	  outparam->get_pcd_lut1pg >>= PCD_PSIZE_16_SHIFT;
	  break;
	case 12:
	  outparam->get_pcd_lut1pg >>= PCD_PSIZE_12_SHIFT;
	  break;
	case 10:
	  outparam->get_pcd_lut1pg >>= PCD_PSIZE_10_SHIFT;
	  break;
	case 8:
	  outparam->get_pcd_lut1pg >>= PCD_PSIZE_8_SHIFT;
	  break;
	}
      break;

      /* write 16 bit input LUT1 (bit 16-31) page number */
    case HIOC_PCD_SET_LUT1PG:
      switch (pcd_get_lut1psz (iti))
	{
	case 16:
	  if (inparam->set_pcd_lut1pg & ~PCD_PSIZE_16_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut1pg <<= PCD_PSIZE_16_SHIFT;
	  break;
	case 14:
	  if (inparam->set_pcd_lut1pg & ~PCD_PSIZE_14_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut1pg <<= PCD_PSIZE_16_SHIFT;
	  break;
	case 12:
	  if (inparam->set_pcd_lut1pg & ~PCD_PSIZE_12_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut1pg <<= PCD_PSIZE_12_SHIFT;
	  break;
	case 10:
	  if (inparam->set_pcd_lut1pg & ~PCD_PSIZE_10_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut1pg <<= PCD_PSIZE_10_SHIFT;
	  break;
	case 8:
	  if (inparam->set_pcd_lut1pg & ~PCD_PSIZE_8_MASK)
	    {
	      error = -EINVAL;
	      goto ioctl_error;
	    }
	  inparam->set_pcd_lut0pg <<= PCD_PSIZE_8_SHIFT;
	  break;
	}
      CHECK_FOR_EIO (pcd_set_lut1pg, inparam->set_pcd_lut1pg);
      break;

      /* read 16 bit input LUT1 (bit 16-31) */
    case HIOC_PCD_GET_LUT1:
      pcd_get_lut (iti, '1', &outparam->get_pcd_lut1);
      break;

      /* write 16 bit input LUT1 (bit 16-31) */
    case HIOC_PCD_SET_LUT1:
      if (pcd_set_lut (iti, '1', &inparam->set_pcd_lut1))
	{
	  iti_error (iti, "pcd_set_lut1");
	  error = -EIO;
	  goto ioctl_error;
	}
      break;
      
      /* linearize lookup table */
    case HIOC_PCD_SET_LUT_LIN:
      if (pcd_lin_lut (iti, '0', ITI_MODE_LINEAR))
	{
	  iti_error (iti, "pcd_lin_lut0");
	  error = -EIO;
	  goto ioctl_error;
	}
      if (pcd_lin_lut (iti, '1', ITI_MODE_LINEAR))
	{
	  iti_error (iti, "pcd_lin_lut1");
	  error = -EIO;
	  goto ioctl_error;
	}
      break;
      
      /* normalize output to 8 bit (from 8, 10, 12, 16 bit input) */
    case HIOC_PCD_SET_NORM_TO8:
      /* the fpga have to be initialized already right */
      if (iti->get_pixsz (iti) != 1)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      if (pcd_lin_lut (iti, '0', ITI_MODE_NORM_TO8))
	{
	  iti_error (iti, "pcd_lin_lut0");
	  error = -EIO;
	  goto ioctl_error;
	}
      if (pcd_lin_lut (iti, '1', ITI_MODE_NORM_TO8))
	{
	  iti_error (iti, "pcd_lin_lut1");
	  error = -EIO;
	  goto ioctl_error;
	}
      /* set pixelsize an recalculate buffersizes */
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      break;

      /* normalize output to 16 bit (from 8, 10, 12, 16 bit input) */
    case HIOC_PCD_SET_NORM_TO16:
      /* the fpga have to be initialized already right */
      if (iti->get_pixsz (iti) != 2)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      if (pcd_lin_lut (iti, '0', ITI_MODE_NORM_TO16))
	{
	  iti_error (iti, "pcd_lin_lut0");
	  error = -EIO;
	  goto ioctl_error;
	}
      if (pcd_lin_lut (iti, '1', ITI_MODE_NORM_TO16))
	{
	  iti_error (iti, "pcd_lin_lut1");
	  error = -EIO;
	  goto ioctl_error;
	}
      /* set pixelsize an recalculate buffersizes */
      if ((error = iti_fmt_init (iti)))
	goto ioctl_error;
      break;

      /* set the exsync mode to free running OR config file default */
    case HIOC_PCD_SET_EXFREE:
      /* find first board */
      for (pcdig = board; pcdig >= 0; pcdig--)
	if (itifg[pcdig]->device_id == ITI_PCDIG &&
	    pcd_get_driveext (itifg[pcdig]) == PCD_ENABLE)
	  break;
      if (pcdig < 0)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      ITI_PRINT1A("    exfree mode : %s\n", inparam->set_pcd_exfree ?
		  "TRUE" : "FALSE");
      if (inparam->set_pcd_exfree)
	{
	  exsclk = pcd_get_exsclk (itifg[pcdig]);
	  exstime = pcd_get_exstime (itifg[pcdig]);
	  switch (exsclk)
	    {
	    case PCD_EXSCLK_3_496_k:
	      exstime <<= 12;
	      break;
	    case PCD_EXSCLK_13_982_k:
	      exstime <<= 10;
	      break;
	    case PCD_EXSCLK_55_930_k:
	      exstime <<= 8;
	      break;
	    case PCD_EXSCLK_223_719_k:
	      exstime <<= 6;
	      break;
	    case PCD_EXSCLK_894_875_k:
	      exstime <<= 4;
	      break;
	    case PCD_EXSCLK_3_580_M:
	      exstime <<= 2;
	      break;
	    case PCD_EXSCLK_7_159_M:
	      exstime <<= 1;
	      break;
	    case PCD_EXSCLK_14_318_M:
	      break;
	    default:
	      error = -EIO;
	      goto ioctl_error;
	    }
	  exstime_convert = exstime;
	  exstime_convert *= ITI_DLY_BASE;
	  pcd_set_exsmd (itifg[pcdig], PCD_EXSYNC_FREE_RUN);
	  /* set the timeout at ALL pcdig's */
	  for (pcdig = boards - 1; pcdig >= 0; pcdig--)
	    if (itifg[pcdig]->device_id == ITI_PCDIG)
	      itifg[pcdig]->cam_expire = (int)((exstime_convert / 1000000000) *
					       (float)HZ * 1.25);
	  ITI_PRINT1A("    cam_expire: %ld\n", iti->cam_expire);
	}
      else
	{
	  pcd_set_exsmd (itifg[pcdig], iti->dev.pcd_cam.exsmd);
	  if (pcd_get_exsmd (itifg[pcdig]) != PCD_EXSYNC_FREE_RUN)
	    {
	      /* set the timeout at ALL pcdig's */
	      for (pcdig = boards - 1; pcdig >= 0; pcdig--)
		itifg[pcdig]->cam_expire = ITI_WAIT_TRG;
	      ITI_PRINT1A("    cam_expire: %ld\n", iti->cam_expire);
	    }
	}
      /* set the mode synchronized */
      iti->opmode |= ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (|SYNC)\n", iti->opmode);
      ITI_PRINT4("    status: sleep\n");
      if ((error = iti_sleep (iti, iti->cam_expire, ITI_CAMERA)))
	goto ioctl_error;
      iti->opmode &= ~ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (&~SYNC)\n", iti->opmode);
      break;

      /* get the exposure time */
    case HIOC_PCD_GET_EXTIME:
      /* find first board */
      for (pcdig = board; pcdig >= 0; pcdig--)
	if (itifg[pcdig]->device_id == ITI_PCDIG &&
	    pcd_get_driveext (itifg[pcdig]) == PCD_ENABLE)
	  break;
      if (pcdig < 0)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      exsclk = pcd_get_exsclk (itifg[pcdig]);
      exstime = pcd_get_exstime (itifg[pcdig]) -
	pcd_get_emidpnt (itifg[pcdig]);
      switch (exsclk)
	{
	case PCD_EXSCLK_3_496_k	:
	  exstime <<= 12;
	  break;
	case PCD_EXSCLK_13_982_k:
	  exstime <<= 10;
	  break;
	case PCD_EXSCLK_55_930_k:
	  exstime <<= 8;
	  break;
	case PCD_EXSCLK_223_719_k:
	  exstime <<= 6;
	  break;
	case PCD_EXSCLK_894_875_k:
	  exstime <<= 4;
	  break;
	case PCD_EXSCLK_3_580_M:
	  exstime <<= 2;
	  break;
	case PCD_EXSCLK_7_159_M:
	  exstime <<= 1;
	  break;
	case PCD_EXSCLK_14_318_M:
	  break;
	default:
	  error = -EIO;
	  goto ioctl_error;
	}
      exstime_convert = exstime;
      exstime_convert *= ITI_DLY_BASE;
      outparam->get_pcd_extime.tv_sec =
	(time_t)(exstime_convert / 1000000000.0);
      outparam->get_pcd_extime.tv_nsec =
	(time_t)(exstime_convert - ((double)outparam->get_pcd_extime.tv_sec *
				   1000000000));
      break;
      
      /* set the exposure time */
      case HIOC_PCD_SET_EXTIME:
      ITI_PRINT1A("    shuttertime : %lds/%ldns.\n",
		  inparam->set_pcd_extime.tv_sec,
		  inparam->set_pcd_extime.tv_nsec);
      exstime_convert = (double)inparam->set_pcd_extime.tv_sec * 1000000000 +
	(double)inparam->set_pcd_extime.tv_nsec +
	iti->dev.pcd_cam.rotime * 1000;
      if (exstime_convert > PCD_DLY_MAX || exstime_convert < PCD_DLY_MIN)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      /* find first board */
      for (pcdig = board; pcdig >= 0; pcdig--)
	if (itifg[pcdig]->device_id == ITI_PCDIG &&
	    pcd_get_driveext (itifg[pcdig]) == PCD_ENABLE)
	  break;
      if (pcdig < 0)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      midpnt_convert = itifg[pcdig]->dev.pcd_cam.rotime * 1000;
      exstime_convert /= ITI_DLY_BASE, midpnt_convert /= ITI_DLY_BASE;
      exstime = exstime_convert, midpnt = midpnt_convert;
      if (exstime > (1 << 10) * 65537)
	exsclk = PCD_EXSCLK_3_496_k, exstime >>= 12, midpnt >>= 12;
      else
	if (exstime > (1 << 8) * 65537)
	  exsclk = PCD_EXSCLK_13_982_k, exstime >>= 10, midpnt >>= 10; 
	else
	  if (exstime > (1 << 6) * 65537)
	    exsclk = PCD_EXSCLK_55_930_k, exstime >>= 8, midpnt >>= 8; 
	  else
	    if (exstime > (1 << 4) * 65537)
	      exsclk = PCD_EXSCLK_223_719_k, exstime >>= 6, midpnt >>= 6; 
	    else
	      if (exstime > (1 << 2) * 65537)
		exsclk = PCD_EXSCLK_894_875_k, exstime >>= 4, midpnt >>= 4; 
	      else
		if (exstime > (1 << 1) * 65537)
		  exsclk = PCD_EXSCLK_3_580_M, exstime >>= 2, midpnt >>= 2; 
		else
		  if (exstime > 65537)
		exsclk = PCD_EXSCLK_7_159_M, exstime >>= 1, midpnt >>= 1; 
      /* update exstime to the new values without interrupts! */
      iti_irq_lock (itifg[pcdig]);
      new_exsclk = exsclk;
      new_exstime = exstime;
      new_midpnt = midpnt;
      iti_irq_unlock (itifg[pcdig]);
      /* adjust the timeout at ALL pcdig's */
      exstime_convert *= ITI_DLY_BASE;
      if (pcd_get_exsmd (itifg[pcdig]) == PCD_EXSYNC_FREE_RUN)
	{
	  for (pcdig = boards - 1; pcdig >= 0; pcdig--)
	    if (itifg[pcdig]->device_id == ITI_PCDIG)
	      itifg[pcdig]->cam_expire = (int)((exstime_convert / 1000000000) *
					       (float)HZ * 1.25);
	  ITI_PRINT1A("    cam_expire: %d %ld\n", pcdig, iti->cam_expire);
	}
      /* update now / update in vb-irq */
      if (iti->opmode & ITI_MODE_GRAB)
	{
	  ITI_PRINT4("    opmode: GRAB\n");
	  need_update = TRUE;
	}
      else
	{
	  ITI_PRINT4("    opmode: !GRAB\n");	  
	  pcd_update_exsync (iti);
	}
      break;

      /* get camera mode */
    case HIOC_PCD_GET_CAMMD:
      outparam->get_pcd_cammd = pcd_get_cammd (iti);
      break;

      /* set camera mode */
    case HIOC_PCD_SET_CAMMD:
      ITI_PRINT1A("    cameramode : %d.\n", inparam->set_pcd_cammd);
      if (inparam->set_pcd_cammd < PCD_CAMMD_MIN ||
	  inparam->set_pcd_cammd > PCD_CAMMD_MAX)
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      /* if four cameras are connected, only bit 0 is valid! */
      if (pcd_get_cam3enb (iti) == PCD_ENABLE &&
	  (inparam->set_pcd_cammd != 0 || inparam->set_pcd_cammd != 0))
	{
	  error = -EINVAL;
	  goto ioctl_error;
	}
      CHECK_FOR_EIO (pcd_set_cammd, inparam->set_pcd_cammd);
      break;
      
    /* Trigger modes. *** PCD_DISABLE is probably the only sensible value for
       some of the sync modes. Check? */
    case HIOC_PCD_GET_STMODE:
      outparam->get_pcd_tmode = pcd_get_stmode (iti);
      break;

    case HIOC_PCD_SET_STMODE:
      ITI_PRINT1A("    start trigger mode : %d.\n", inparam->set_pcd_tmode);
      pcd_set_stmode (iti, inparam->set_pcd_tmode);
      break;

    case HIOC_PCD_GET_ETMODE:
      outparam->get_pcd_tmode = pcd_get_etmode (iti);
      break;

    case HIOC_PCD_SET_ETMODE:
      ITI_PRINT1A("    end trigger mode : %d.\n", inparam->set_pcd_tmode);
      pcd_set_etmode (iti, inparam->set_pcd_tmode);
      break;

    case HIOC_PCD_GET_FCNT:
      outparam->get_pcd_fcnt = pcd_get_fcnt (iti);
      break;

    case HIOC_PCD_SET_FCNT:
      ITI_PRINT1A("    frame count : %d.\n", inparam->set_pcd_fcnt);
      pcd_set_fcnt (iti, inparam->set_pcd_fcnt);
      break;

    default:
      error = -EINVAL;
      goto ioctl_error;
    }
  
ioctl_done:
  ITI_PRINT0A("< iti_ioctl %d: OK\n", board);
  return OK;
  
ioctl_error:
  ITI_PRINT0A("< iti_ioctl %d: error %d\n", board, error);
  return error;
}

